"""Blocks to help with actors"""

import serge.actor
import serge.engine
import serge.actor

class MissingBehaviour(Exception): """Could not locate the behaviour"""
class DuplicateBehaviour(Exception): """The behaviour was already recorded"""
class BehaviourNotPaused(Exception): """The behaviour was not paused"""
class BehaviourAlreadyPaused(Exception): """The behaviour was already paused"""



class ScreenActor(serge.actor.CompositeActor):
    """An actor to represent the logic associated with a screen of the game
    
    This actor is useful when encapsulating the logic associated with a specific
    screen in the game. The actor has useful properties and methods that
    make it easy to manage the logic.
    
    """
    
    def __init__(self, *args, **kw):
        """Initialise the ScreenActor"""
        super(ScreenActor, self).__init__(*args, **kw)
        
    def addedToWorld(self, world):
        """The actor was added to the world"""
        super(ScreenActor, self).addedToWorld(world)
        self.world = world
        self.engine = serge.engine.CurrentEngine()
        self.keyboard = self.engine.getKeyboard()
        self.mouse = self.engine.getMouse()
        

class RepeatedVisualActor(serge.actor.Actor):
    """An actor that shows multiple copies of a visual representation
    
    This actor is useful for showing the number of lives or missiles
    etc in a game.
    
    """
    
    def __init__(self, tag, name=None, repeat=5, spacing=10, orientation='horizontal'):
        """Initialise the RepeatedVisualActor"""
        super(RepeatedVisualActor, self).__init__(tag, name)
        self._repeat = repeat
        self._spacing = spacing
        self._current = repeat
        self._orientation = orientation

    def _resetVisual(self):
        """Reset the visual item on the center point
        
        We need to override this because our size is not determined by our visual
        
        """
        #
        # Adjust our location so that we are positioned and sized appropriately
        cx, cy, _, _ = self.getSpatialCentered()
        #
        if self._orientation == 'horizontal':
            self.setSpatialCentered(cx, cy, 
                self._visual.width + self._spacing*(self._repeat-1), self._visual.height)
        else:
            self.setSpatialCentered(cx, cy, 
                self._visual.width, self._visual.height + self._spacing*(self._repeat-1))
        #
        # Here is a hack - sometimes the visual width changes and we want to update our width
        # so we let the visual know about us so it can update our width. This is almost 
        # certainly the wrong thing to do, but we have some tests in there so hopefully
        # the right thing becomes obvious later!
        self._visual._actor_parent = self
        
    def renderTo(self, renderer, interval):
        """Render ourself to the given renderer"""
        if self._visual:
            ox, oy = renderer.camera.getRelativeLocation(self)
            if self.layer:
                for i in range(self._current):
                    if self._orientation == 'horizontal':
                        x, y = (ox + i*self._spacing, oy)
                    else:
                        x, y = (ox, oy + i*self._spacing)
                    self._visual.renderTo(interval, renderer.getLayer(self.layer).getSurface(), (x, y))

    def reduceRepeat(self, amount=1):
        """Reduce the repeat by a certain amount"""
        self.setRepeat(self._current - amount)
        
    def increaseRepeat(self, amount=1):
        """Increase the repeat by a certain amount"""
        self.setRepeat(self._current + amount)
        
    def getRepeat(self):
        """Return the current repeat"""
        return self._current
    
    def setRepeat(self, value):
        """Set the current repeat"""
        self._current = value
        #
        # Reset the visual size
        ox, oy, w, h = self.getSpatial()
        if self._orientation == 'horizontal':
            w = self._visual.width + self._spacing*(self._current-1)
        else:
            h = self._visual.height + self._spacing*(self._current-1)
        self.setSpatial(ox, oy, w, h)
        self.log.debug('New spatial = %s' % self.getSpatial())
        
    def resetRepeat(self):
        """Reset the repeat to the initial value"""
        self.setRepeat(self._repeat)
        
        
class FormattedText(serge.actor.Actor):
    """A text display that can be formatted"""
    
    def __init__(self, tag, name, format, colour, font_name='DEFAULT', font_size=12, justify='center', **kw):
        """Initialise the text"""
        super(FormattedText, self).__init__(tag, name)
        self.visual = serge.visual.Text('', colour, font_name, font_size, justify)
        self.format = format
        self.values = kw
        self.updateText()
        
    def updateText(self):
        """Update our text"""
        self.visual.setText(self.format % self.values)
    
    def setValue(self, name, value):
        """Set the value"""
        self.values[name] = value
        self.updateText()
            
    def getValue(self, name):
        """Get the values"""
        return self.values[name]

class NumericText(FormattedText):
    """A helper actor to display some text with a single number in there"""
    
    def __init__(self, *args, **kw):
        """Initialise the text"""
        super(NumericText, self).__init__(*args, **kw)
        
    def updateText(self):
        """Update our text"""
        self.visual.setText(self.format % self.values['value'])

    @property
    def value(self): return self.getValue('value')
    @value.setter
    def value(self, v): self.setValue('value', v)

    
class StringText(FormattedText):
    """A helper actor to display some text with text in there"""
    
    def __init__(self, tag, name, text, format='%s', colour=(255, 255, 255), font_name='DEFAULT', font_size=12, justify='center'):
        """Initialise the text"""
        super(StringText, self).__init__(tag, name, format, colour, font_name, font_size, justify, value=text)
        
    def updateText(self):
        """Update our text"""
        self.visual.setText(self.values['value'])

    @property
    def value(self): return self.getValue('value')
    @value.setter
    def value(self, v): self.setValue('value', v)

