"""Classes the implement behaviours"""

import pygame

import serge.actor
import serge.engine
import serge.common
import serge.events

class MissingBehaviour(Exception): """Could not locate the behaviour"""
class DuplicateBehaviour(Exception): """The behaviour was already recorded"""
class BehaviourNotPaused(Exception): """The behaviour was not paused"""
class BehaviourAlreadyPaused(Exception): """The behaviour was already paused"""


### Main classes for implementing behaviours ###


class BehaviourRecord(object):
    """Represents a record of a requested behaviour"""
    
    def __init__(self, actor, behaviour, name):
        """Initialise the record"""
        self._actor = actor
        self._behaviour = behaviour
        self._name = name
        self._running = True
        if isinstance(behaviour, Behaviour):
            self._behaviour.record = self
                
    def performBehaviour(self, interval, world):
        """Perform the actual behaviour"""
        if self._running:
            self._behaviour(world, self._actor, interval)
        
    def pause(self):
        """Pause the behaviour"""
        if self._running:
            self._running = False
        else:
            raise BehaviourAlreadyPaused('The behaviour %s was not paused' % self)

    def restart(self):
        """Restart the behaviour"""
        if not self._running:
            self._running = True
        else:
            raise BehaviourNotPaused('The behaviour %s was already running' % self)

    def matches(self, actor, name):
        """Return True if this behaviour matches the actor and name"""
        return self._actor == actor and self._name == name

    def matchesName(self, name):
        """Return True if this behaviour matches the name"""
        return self._name == name

    def _getId(self):
        """Return an id to identify the behaviour"""
        nice_name = 'the game' if self._actor is None else self._actor.getNiceName()
        return '%s-%s' % (self._name, nice_name)
        
                   
class BehaviourManager(serge.actor.Actor):
    """Manages the behaviour of multiple actors in a world"""
    
    def __init__(self, *args, **kw):
        """Initialise the manager"""
        super(BehaviourManager, self).__init__(*args, **kw)
        self._behaviours = {}
        
    def assignBehaviour(self, actor, behaviour, name):
        """Assign a behaviour to an actor"""
        record = BehaviourRecord(actor, behaviour, name)
        if record._getId() in self._behaviours:
            raise DuplicateBehaviour('The behaviour %s,%s was already recorded' % (actor, behaviour))
        nice_name = 'the game' if actor is None else actor.getNiceName()
        self.log.info('Assigned behaviour %s to %s' % (behaviour, nice_name))
        self._behaviours[record._getId()] = record
        return record
        
    def removeBehaviour(self, record):
        """Remove a particular behaviour"""
        try:
            del(self._behaviours[record._getId()])
        except KeyError:
            raise MissingBehaviour('The behaviour %s was not found' % record)

    def removeBehaviourByName(self, actor, name):
        """Remove the named behaviour for an actor based on its name"""
        for record in self._behaviours.values():
            if record.matches(actor, name):
                self.removeBehaviour(record)
                return
        raise MissingBehaviour('Not behaviour named "%s" found for actor "%s"' % (name, actor.getNiceName()))

    def removeBehavioursByName(self, name):
        """Remove the named behaviour for all actors based on a name"""
        for record in self._behaviours.values():
            if record.matchesName(name):
                self.removeBehaviour(record)

    def updateActor(self, world, interval):
        """Perform all our behaviours"""
        for behaviour in self._behaviours.values():
            behaviour.performBehaviour(world, interval)
            
    def pauseBehaviours(self, name):
        """Pause all behaviours with the name"""
        for record in self._behaviours.values():
            if record.matchesName(name):
                record.pause()

    def restartBehaviours(self, name):
        """Restart all behaviours with the name"""
        for record in self._behaviours.values():
            if record.matchesName(name):
                record.restart()
        

class Behaviour(serge.common.Loggable):
    """Base class for all behaviours"""
    
    def __init__(self):
        """Initialise the behaviour"""
        self.addLogger()
        self.active = True
        
    def __call__(self, world, actor, interval):
        """Execute the behaviour"""
        raise NotImplementedError
            
            
### Some useful behaviours ###

class MoveTowardsPoint(Behaviour):
    """Move an actor towards a point"""
    
    def __init__(self, point, x_speed=1, y_speed=1):
        """Initialise the behaviour
        
        You can limit the directions that will be moved by setting
        x_speed or y_speed to be 0.
        
        """
        super(MoveTowardsPoint, self).__init__()
        self._target = point
        self._x_speed = x_speed
        self._y_speed = y_speed
        
    def __call__(self, world, actor, interval):
        """Do the movement"""
        dx = (self._target[0] - actor.x)
        if dx > 0:
            actor.x += min(dx, self._x_speed)
        elif dx < 0:
            actor.x -= min(abs(dx), self._x_speed)
        #   
        dy = (self._target[1] - actor.y)
        if dy > 0:
            actor.y += min(dy, self._y_speed)
        elif dy < 0:
            actor.y -= min(abs(dy), self._y_speed)


class MoveTowardsActor(Behaviour):
    """Move an actor towards another actor"""
    
    def __init__(self, actor, x_speed=1, y_speed=1):
        """Initialise the behaviour
        
        You can limit the directions that will be moved by setting
        x_speed or y_speed to be 0.
        
        """
        super(MoveTowardsActor, self).__init__()
        self._target = actor
        self._x_speed = x_speed
        self._y_speed = y_speed
        
    def __call__(self, world, actor, interval):
        """Do the movement"""
        dx = (self._target.x - actor.x)
        if dx > 0:
            actor.x += min(dx, self._x_speed)
        elif dx < 0:
            actor.x -= min(abs(dx), self._x_speed)
        #   
        dy = (self._target.y - actor.y)
        if dy > 0:
            actor.y += min(dy, self._y_speed)
        elif dy < 0:
            actor.y -= min(abs(dy), self._y_speed)
     
        
class AvoidActor(Behaviour):
    """Move away from an actor until you reach a certain distance"""
    
    def __init__(self, actor, x_speed=1, y_speed=1, distance=10):
        """Initialise the behaviour
        
        You can limit the directions that will be moved by setting
        x_speed or y_speed to be 0.
        
        """
        super(AvoidActor, self).__init__()
        self._target = actor
        self._distance = distance
        self._x_speed = x_speed
        self._y_speed = y_speed
        
    def __call__(self, world, actor, interval):
        """Do the movement if we are within range"""
        #
        # Check if we are outside of the given distance
        dx = (self._target.x - actor.x)
        dy = (self._target.y - actor.y)
        if (dx**2 + dy**2) > self._distance**2:
            return
        #
        if dx > 0:
            actor.x -= min(dx, self._x_speed)
        elif dx < 0:
            actor.x += min(abs(dx), self._x_speed)
        #   
        if dy > 0:
            actor.y -= min(dy, self._y_speed)
        elif dy < 0:
            actor.y += min(abs(dy), self._y_speed)


class TwoOptions(Behaviour):
    """A behaviour that chooses between two optional behaviours"""
    
    def __init__(self, b1, b2, arg, selector):
        """Initialise the behaviour
        
        b1 and b2 are the two behaviours that we want to choose between.
        The selector function will be called by passing 'arg' to it. The
        selector should return True to activate b1, False for b2.
        
        """
        super(TwoOptions, self).__init__()
        self._b1 = b1
        self._b2 = b2
        self._arg = arg
        self._selector = selector
        
    def __call__(self, world, actor, interval):
        """Do the appropriate behaviour"""
        if self._selector(self._arg):
            self._b1(world, actor, interval)
        else:
            self._b2(world, actor, interval)
            
class Optional(TwoOptions):
    """A behaviour that is turned on and off by an option"""
    
    def __init__(self, behaviour, arg, selector):
        """Initialise the behaviour
        
        The selector function will be called by passing 'arg' to it. The
        selector should return True to activate the behaviour
        
        """
        super(Optional, self).__init__(behaviour, lambda a,b,c:None, arg, selector)         


### Control behaviours ###


class KeyboardNSEW(Behaviour):
    """Move an actor in ordinal directions according to the keyboard
    
    Set the n, s, e and w to the keys you want to move the actor. If you
    do not want any motion then set that direction to None. Set the speed
    to be the amount to move per keypress.
    
    """
    
    def __init__(self, speed, n=pygame.K_UP, s=pygame.K_DOWN, e=pygame.K_RIGHT, w=pygame.K_LEFT):
        """Initialise the behaviour"""
        super(KeyboardNSEW, self).__init__()
        self._speed = speed
        self._n = n
        self._s = s
        self._e = e
        self._w = w
        self._keyboard = serge.engine.CurrentEngine().getKeyboard()
        
    def __call__(self, world, actor, interval):
        """Perform the motion"""
        dy = dx = 0
        if self._n and self._keyboard.isDown(self._n):
            dy = -self._speed
        if self._s and self._keyboard.isDown(self._s):
            dy = +self._speed
        if self._e and self._keyboard.isDown(self._e):
            dx = +self._speed
        if self._w and self._keyboard.isDown(self._w):
            dx = -self._speed
        actor.move(dx, dy)        


class KeyboardNSEWToVectorCallback(Behaviour):
    """Calls a method with direction vector from ordinal directions according to the keyboard
    
    Set the n, s, e and w to the keys you want to move the actor. If you
    do not want any motion then set that direction to None. Set the speed
    to be the amount to move per keypress.
    
    This is useful when you want to move an object but you need to do
    some preprocessing first. This behaviour will allow you to capture the
    keypresses.
    
    """
    
    def __init__(self, method, event=serge.events.E_KEY_CLICKED, speed=1, 
            n=pygame.K_UP, s=pygame.K_DOWN, e=pygame.K_RIGHT, w=pygame.K_LEFT):
        """Initialise the behaviour"""
        super(KeyboardNSEWToVectorCallback, self).__init__()
        self._method = method
        self._event_method = 'isDown' if event == serge.events.E_KEY_DOWN else 'isClicked'
        self._speed = speed
        self._n = n
        self._s = s
        self._e = e
        self._w = w
        self._keyboard = serge.engine.CurrentEngine().getKeyboard()
        
    def __call__(self, world, actor, interval):
        """Perform the motion"""
        dy = dx = 0
        if self._n and getattr(self._keyboard, self._event_method)(self._n):
            dy = -self._speed
        if self._s and getattr(self._keyboard, self._event_method)(self._s):
            dy = +self._speed
        if self._e and getattr(self._keyboard, self._event_method)(self._e):
            dx = +self._speed
        if self._w and getattr(self._keyboard, self._event_method)(self._w):
            dx = -self._speed
        if dx or dy:
            self._method((dx, dy))

        
class KeyboardQuit(Behaviour):
    """Quit the game based on a keypress"""
    
    def __init__(self, key=pygame.K_ESCAPE):
        """Initialise the behaviour"""
        super(KeyboardQuit, self).__init__()
        self._key = key
        self._keyboard = serge.engine.CurrentEngine().getKeyboard()
        
    def __call__(self, world, actor, interval):
        """Check for the quit key"""
        if self._keyboard.isDown(self._key):
            self.log.info('Quit key pressed')
            serge.engine.CurrentEngine().stop()
            
            
class TimedCallback(Behaviour):
    """A callback that gets called at a certain interval"""
    
    def __init__(self, interval, callback):
        """Initialise the behaviour"""
        self.interval = interval
        self.callback = callback
        self.timer = 0.0
        self._number_calls = 0
        
    def __call__(self, world, actor, interval):
        """Check for the interval"""
        self.timer += interval
        if self.timer >= self.interval:
            self.timer -= self.interval            
            self.callback(world, actor, interval)
            self._number_calls += 1

class TimedOneshotCallback(TimedCallback):
    """A callback that gets called once and only once at an interval"""
    
    def __call__(self, world, actor, interval):
        """Check for the interval"""
        super(TimedOneshotCallback, self).__call__(world, actor, interval)
        if self._number_calls:
            self.record.pause()
            
