"""Useful blocks for visual rendering"""

import pygame

import serge.visual

class InvalidSprite(Exception): """The selected sprite was not valid"""
class InvalidParameters(Exception): """The parameters for the shape were not valid"""

### Simple shapes ###


class Rectangle(serge.visual.Drawing):
    """A rectangle"""
    
    def __init__(self, (w, h), colour, stroke_width=0, stroke_colour=None):
        """Initialise the rectangle"""
        super(Rectangle, self).__init__()
        self.width, self.height = w, h
        self.colour = colour
        self.stroke_colour = stroke_colour 
        self.stroke_width = stroke_width
        if (stroke_width and stroke_colour) and (stroke_width > min(w, h)/2 or stroke_width < 0):
            raise InvalidParameters('Stroking width must be > 0 and less than half the smallest side (%d)' % stroke_width)
        
    def renderTo(self, milliseconds, surface, (x, y)):
        """Render to a surface"""
        if self.stroke_colour and self.stroke_width:
            #
            # If we are stroking then draw a full size rectange with the stroke colour and
            # then a smaller one with the actual fill colour
            pygame.draw.rect(surface, self.stroke_colour, (x, y, self.width, self.height))
            pygame.draw.rect(surface, self.colour, (x+self.stroke_width, y+self.stroke_width, 
                self.width-2*self.stroke_width, self.height-2*self.stroke_width))
        else:
            pygame.draw.rect(surface, self.colour, (x, y, self.width, self.height), self.stroke_width)
        
        
class Circle(serge.visual.Drawing):
    """A circle"""
    
    def __init__(self, radius, colour, stroke_width=0, stroke_colour=None):
        """Initialise the circle"""
        super(Circle, self).__init__()
        self.radius = radius
        self.colour = colour
        self.stroke_colour = stroke_colour 
        self.stroke_width = stroke_width
        if (stroke_width and stroke_colour) and (stroke_width > radius or stroke_width < 0):
            raise InvalidParameters('Stroking width must be > 0 and less than the radius (%d)' % stroke_width)
        
    def renderTo(self, milliseconds, surface, (x, y)):
        """Render to a surface"""
        if self.stroke_colour and self.stroke_width:
            #
            # If we are stroking then draw a full size circle with the stroke colour and
            # then a smaller one with the actual fill colour
            pygame.draw.circle(surface, self.stroke_colour, (int(x), int(y)), self.radius)
            pygame.draw.circle(surface, self.colour, (int(x), int(y)), self.radius-self.stroke_width)
        else:
            pygame.draw.circle(surface, self.colour, (int(x), int(y)), self.radius, self.stroke_width)
           
    def setAngle(self, angle):
        """Set the angle
        
        Pass through as this is a circle!"""
        pass

### Shapes with text ###

class RectangleText(serge.visual.Drawing):
    """A rectangle with some text on it"""
    
    def __init__(self, text, text_colour, rect_dimensions, rect_colour, font_size=12, font_name='DEFAULT', stroke_width=0, stroke_colour=None, justify='center'):
        """Initialise the drawing"""
        self.text_visual = serge.visual.Text(text, text_colour, font_name, font_size, justify)
        self.rect_visual = Rectangle(rect_dimensions, rect_colour, stroke_width, stroke_colour)
        self.width = self.rect_visual.width
        self.height = self.rect_visual.height
        
    def renderTo(self, milliseconds, surface, (x, y)):
        """Render to a surface"""
        w, h = self.rect_visual.width, self.rect_visual.height
        self.rect_visual.renderTo(milliseconds, surface, (x, y))
        self.text_visual.renderTo(milliseconds, surface, (x+w/2-self.text_visual.width/2, y+h/2-self.text_visual.height/2))
        
class CircleText(serge.visual.Drawing):
    """A circle with some text on it"""
    
    def __init__(self, text, text_colour, radius, circle_colour, font_size=12, font_name='DEFAULT', stroke_width=0, stroke_colour=None, justify='center'):
        """Initialise the drawing"""
        self.text_visual = serge.visual.Text(text, text_colour, font_name, font_size, justify)
        self.circle_visual = Circle(radius, circle_colour, stroke_width, stroke_colour)
        
    def renderTo(self, milliseconds, surface, (x, y)):
        """Render to a surface"""
        self.circle_visual.renderTo(milliseconds, surface, (x, y))
        self.text_visual.renderTo(milliseconds, surface, 
            (x-self.text_visual.width/2, 
             y-self.text_visual.height/2))

    def getSize(self):
        """Return the size of the drawing"""
        return [self.radius]
        
### Sprite with text ###

class SpriteText(serge.visual.Sprite):
    """A sprite with some text on it"""

    def __init__(self, text, text_colour, sprite_name, font_size=12, font_name='DEFAULT', stroke_width=0, stroke_colour=None, justify='center'):
        """Initialise the drawing"""
        super(SpriteText, self).__init__()
        self.text_visual = serge.visual.Text(text, text_colour, font_name, font_size, justify)
        sprite = serge.visual.Register.getItem(sprite_name)
        self.setImage(sprite.raw_image, (sprite.width, sprite.height), sprite.framerate, sprite.running)
        
    def renderTo(self, milliseconds, surface, (x, y)):
        """Render to a surface"""
        super(SpriteText, self).renderTo(milliseconds, surface, (x, y))
        self.text_visual.renderTo(milliseconds, surface, 
                (x+self.width/2-self.text_visual.width/2, 
                 y+self.height/2-self.text_visual.height/2))

    def setText(self, text):
        """Set the text"""
        self.text_visual.setText(text)

    @property
    def text(self): return self.text_visual.text

class TextToggle(SpriteText):
    """A sprite text item that has multiple cells and can be used as a toggle
    
    You can set the cells directly of use On=0 and Off=1.
    
    """
    
    def __init__(self, *args, **kw):
        """Initialise the  TextToggle"""
        super(TextToggle, self).__init__(*args, **kw)
        #
        # Reality check - the underlying sprite must have at least two cells
        if len(self.cells) <= 1:
            raise InvalidSprite('The selected sprite does not have enough cells. Needs at least two')
        
    def setOn(self):
        """Set to on"""
        self.setCell(0)
        
    def setOff(self):
        """Set to off"""
        self.setCell(1)
        
    def toggle(self):
        """Toggle the state"""
        if self.current_cell == 0:
            self.setOff()
        else:
            self.setOn()
    
    def isOn(self):
        """Return if we are on"""
        return self.current_cell == 0
    
    def isOff(self):
        """Return if we are on"""
        return self.current_cell != 0
    
    
class Toggle(TextToggle):
    """Like a text toggle but with no text"""
    
    def __init__(self, sprite_name):
        """Initialise the toggle"""
        super(Toggle, self).__init__('', (0,0,0,0), sprite_name)



        
