"""Visual effects"""

import pygame
try:
    import Numeric
except ImportError:
    Numeric = None

import serge.visual
import serge.render

if not Numeric:
    ShadowLayer = lambda a, b, c, d: serge.render.Layer(a, b)
else:
    class Shadow(serge.visual.SurfaceDrawing):
        """Creates a shadow from an image"""
        
        def __init__(self, source, colour):
            """Initialise the Shadow"""
            super(Shadow, self).__init__(*source.get_size())
            self._source = source
            self._colour = colour
            self.createShadow()
            
        def createShadow(self):
            """Create the shadow now
            
            Most of the logic here from http://pygame.org/wiki/ShadowEffects
            
            """
            #
            # Create alpha of main image
            image = self._source
            ambience = float(255-self._colour[3])/255
            if image.get_masks()[3] != 0:
                image_alpha = pygame.surfarray.pixels_alpha(image)
                if ambience > 0.0:
                    shadow_alpha = (image_alpha *
                                    (1.0 - ambience)).astype(Numeric.UInt8)
                else:
                    shadow_alpha = image_alpha
            elif image.get_colorkey() is not None:
                image_alpha = pygame.surfarray.array_colorkey(image)
                image.unlock(); image.unlock()  # pygame 1.7 bug (fixed in 1.8).
                surface_alpha = image.get_alpha()
                if surface_alpha is not None:
                    # Do what array_colorkey should have done: use surface alpha!
                    Numeric.minimum(image_alpha, surface_alpha, image_alpha)
                if ambience > 0.0:
                    shadow_alpha = (image_alpha *
                                    (1.0 - ambience)).astype(Numeric.UInt8)
                else:
                    shadow_alpha = image_alpha
            else:
                image_alpha = image.get_alpha()
                if image_alpha is None:
                    image_alpha = 255
                shadow_alpha = int(image_alpha * (1.0 - ambience))
            #
            # Make the shadow
            shadow = image.convert_alpha()
            shading = self.getSurface()
            shading.fill(self._colour)
            pygame.surfarray.pixels_alpha(shading)[...] = image_alpha
            shadow.blit(shading, (0, 0))
            pygame.surfarray.pixels_alpha(shadow)[...] = shadow_alpha
            #
            self.surface = shadow


    class ShadowLayer(serge.render.Layer):
        """A layer that renders with a shadow beneath it"""
        
        def __init__(self, name, order, colour, offset):
            """Initialise the ShadowLayer"""
            super(ShadowLayer, self).__init__(name, order)
            self._offset = offset
            self._colour = colour
            
        def initSurface(self, renderer):
            """Initialise the surface"""
            super(ShadowLayer, self).initSurface(renderer)
            self._shadow = Shadow(self.getSurface(), self._colour)
        
        def render(self, surface):
            """Render to a surface
            
            When rendering to the surface we first create our shadow then
            render this to the surface followed by our normal rendering.
            
            """
            self._shadow.createShadow()
            surface.blit(self._shadow.surface, self._offset)
            super(ShadowLayer, self).render(surface)


class FadingLayer(serge.render.Layer):
    """A layer that you can fade in and out"""

    def __init__(self, name, order):
        """Initialise the layer"""
        super(FadingLayer, self).__init__(name, order)
        self.visibility = 255
    
    def postRender(self):
        """After rendering the surface"""
        v = 255-self.visibility
        if v:
            self.getSurface().fill((v, v, v, v), special_flags=pygame.BLEND_RGBA_SUB)   
         
